#!/usr/bin/env python
# -*- mode: python; coding: koi8-r -*-


import string, sys, os

import gobject, gtk, pango
from gtk import TRUE, FALSE
from gtk import gdk


skins_dir = 'skins'

#PANGO_SCALE = pango.SCALE
PANGO_SCALE = 1024

class PangoLayoutWord:
    pangolayout = None
    width = 0
    index = 0


class OrnamentBook(gtk.DrawingArea):

    keysyms_next = (gtk.keysyms.KP_Page_Down,
                    gtk.keysyms.Page_Down,
                    gtk.keysyms.space,
                    gtk.keysyms.Right)
    keysyms_prev = (gtk.keysyms.KP_Page_Up,
                    gtk.keysyms.Page_Up,
                    gtk.keysyms.Left)

    def __init__(self, parent):


        default_skin = 'golden.jpg'
        skin_file = None
        for d in sys.path:
            f = os.path.join(d, 'ornamentbook', skins_dir, default_skin)
            if os.path.exists(f):
                skin_file = f
                break
        if not skin_file:
            sys.exit('ERROR: can\'t find default skin file %s' % 'golden.jpg')
        self.skin_file = skin_file

        gtk.DrawingArea.__init__(self)

        self._parent = parent
        #self.area = gtk.DrawingArea()

##         self.set_size_request(950, 400)
##         self._width = 950
##         self._height = 400

        gobject.signal_new('position-changed', OrnamentBook,
                           gobject.SIGNAL_RUN_LAST,
                           gobject.TYPE_NONE,
                           (gobject.TYPE_FLOAT,))

        self.set_size_request(600, 400)
        self._width = 600
        self._height = 400

        #self.set_events(gdk.ALL_EVENTS_MASK)
        #self.add_events(gdk.BUTTON_PRESS_MASK|gdk.BUTTON_RELEASE_MASK)
        #self.add_events(gdk.BUTTON_RELEASE_MASK)
        self.add_events(gdk.BUTTON_PRESS_MASK)
        #self.add_events(gdk.BUTTON_PRESS_MASK|gdk.BUTTON_RELEASE_MASK|gdk.KEY_PRESS_MASK|gdk.KEY_RELEASE_MASK)
        #self.set_events(gdk.ALL_EVENTS_MASK)

        self.connect('expose-event', self.area_expose_cb)
        #self.connect('button_release_event', self.button_release_event_cb)
        self.connect('button_press_event', self.button_release_event_cb)
        #self.realize()

        #self._parent.connect('key_press_event', self.key_press_event_cb)
        self._parent.connect('key-release-event', self.key_release_event_cb)

        self.text = ''
        self.lines = []

        self.current_page_line_index = 0
        self.current_page_char_index = 0

        self.next_page_line_index = 0
        self.next_page_char_index = 0

        self.pixmap_area = None
        #self.prev_pixmap_area = None
        #self.next_pixmap_area = None

        self.set_font('Normal')
        self.color = gdk.color_parse('#000000')

        self.top_margin = 10
        self.bottom_margin = 10
        self.left_page_left_margin = 30
        self.left_page_right_margin = 30
        self.right_page_left_margin = 30
        self.right_page_right_margin = 30

        #self.tabs_width = 5
        self.set_tab_width(5)

        return


    def set_skin(self, filename):
        try:
            open(filename)
        except IOError, err:
            sys.stderr.write('ERROR: can\'t open file: %s: %s\n'
                             % (filename, err[1]))
            return
        self.skin_file = filename
        self.pixbuf = gdk.pixbuf_new_from_file(self.skin_file)
        self.cur_pixbuf = self.pixbuf.scale_simple(self._width, self._height,
                                                   gdk.INTERP_BILINEAR)

    def set_text(self, text, line=0, index=0):
        self.text = text #unicode(text, 'utf-8')
        self.lines = self.text.splitlines()
        self.current_page_line_index = line
        self.current_page_char_index = index
        #self.pixmap_area = None
        #self.prev_pixmap_area = None
        #self.next_pixmap_area = None

        gtk.idle_add(self.emit_position_changed)


    def set_font(self, font):
        self.font = pango.FontDescription(font)
        pangolayout = self.create_pango_layout(' ')
        pangolayout.set_font_description(self.font)
        self.space_width = pangolayout.get_extents()[1][2]/PANGO_SCALE
        self.text_height = pangolayout.get_extents()[1][3]/PANGO_SCALE


    def set_tab_width(self, tab_width=None):
        if tab_width is None:
            tab_width = self.tabs_width
        pl = self.create_pango_layout(' '*tab_width)
        width, height = pl.get_pixel_size()

        ta = pango.TabArray(1, TRUE)
        ta.set_tab(0, pango.TAB_LEFT, width)
        self.tab_array = ta
        self.tabs_width = tab_width


    def realize(self):
        gtk.DrawingArea.realize(self)
        self.gc = self._parent.style.fg_gc[gtk.STATE_NORMAL]
        self.pixbuf = gdk.pixbuf_new_from_file(self.skin_file)
        self.cur_pixbuf = self.pixbuf.scale_simple(self._width, self._height,
                                                   gdk.INTERP_BILINEAR)


    def area_expose_cb(self, area, event):
        x, y, w, h, d = self.window.get_geometry()
        if self.pixmap_area and self._width == w and self._height == h:
            self.window.draw_drawable(self.gc, self.pixmap_area,
                                      0, 0, 0, 0, w, h)
        else:
            self.draw_text()


    def draw_text(self):

        x, y, widget_width, widget_height, depth = self.window.get_geometry()

        if not self.pixmap_area or self._width != widget_width or self._height != widget_height:
            self.cur_pixbuf = self.pixbuf.scale_simple(widget_width,
                                                       widget_height,
                                                       gdk.INTERP_BILINEAR)
            self._width = widget_width
            self._height = widget_height

            self.pixmap_area = gdk.Pixmap(None,
                                          widget_width,
                                          widget_height,
                                          depth)
##             self.prev_pixmap_area = gdk.Pixmap(None,
##                                                widget_width,
##                                                widget_height,
##                                                depth)
##             self.next_pixmap_area = gdk.Pixmap(None,
##                                                widget_width,
##                                                widget_height,
##                                                depth)

        #if not self.pixmap_area:
        self._draw(self.pixmap_area,
                   self.current_page_line_index,
                   self.current_page_char_index)

        self.window.draw_drawable(self.gc, self.pixmap_area,
                                  0, 0, 0, 0, widget_width, widget_height)

        #gtk.idle_add(self._draw_fb)

##     def _draw_fb(self):
##         # draw fore page
##         if not self.next_pixmap_area:
##             self._draw(self.next_pixmap_area,
##                        self.next_page_line_index,
##                        self.next_page_char_index)
##         # draw back page
##         if not self.prev_pixmap_area:
##             line_index, index = self.calc_back()
##             self._draw(self.prev_pixmap_area, line_index, index)


    def _draw(self, area_window, line_index, index):

        widget_width = self._width
        widget_height = self._height

        pixmap, mask = self.cur_pixbuf.render_pixmap_and_mask()
        area_window.draw_drawable(self.gc, pixmap,
                                  0, 0, 0, 0, -1, -1)

        text_area_height = widget_height-self.top_margin-self.bottom_margin

        #line_index = self.current_page_line_index
        #index = self.current_page_char_index

        # draw left page
        x_offset = self.left_page_left_margin
        y_offset = self.top_margin
        text_area_width = widget_width/2-self.left_page_left_margin \
                          -self.left_page_right_margin

        if index: # draw tail of line if need
            prev_index = index
            y_offset, index = self.draw_line(area_window,
                                             self.lines[line_index][index:],
                                             x_offset,
                                             y_offset,
                                             text_area_width,
                                             text_area_height)
            if not index:
                line_index += 1
            else:
                index += prev_index

        if y_offset+self.text_height <= self.top_margin+text_area_height:
            for line in self.lines[line_index:]:
                y_offset, index = self.draw_line(area_window,
                                                 line,
                                                 x_offset,
                                                 y_offset,
                                                 text_area_width,
                                                 text_area_height)
                if not index:
                    line_index += 1
                if y_offset+self.text_height > self.top_margin+text_area_height:
                    break

        # draw right page
        x_offset = widget_width/2+self.right_page_left_margin
        y_offset = self.top_margin
        text_area_width = widget_width/2-self.right_page_left_margin \
                          -self.right_page_right_margin

        if index: # draw tail of line if need
            prev_index = index
            y_offset, index = self.draw_line(area_window,
                                             self.lines[line_index][index:],
                                             x_offset,
                                             y_offset,
                                             text_area_width,
                                             text_area_height)
            if not index:
                line_index += 1
            else:
                index += prev_index

        if y_offset+self.text_height <= self.top_margin+text_area_height:
            for line in self.lines[line_index:]:

                y_offset, index = self.draw_line(area_window,
                                                 line,
                                                 x_offset,
                                                 y_offset,
                                                 text_area_width,
                                                 text_area_height)


                if not index:
                    line_index += 1
                if y_offset+self.text_height > self.top_margin+text_area_height:
                    break

        self.next_page_char_index = index
        self.next_page_line_index = line_index

        #area_window.thaw_updates()

        return


    def draw_line(self, area_window, line, x_offset, y_offset,
                  text_area_width, text_area_height):

        pangolayouts_list = self.get_pangolayout_list(line, text_area_width)

        if not pangolayouts_list:
            # empty line
            return y_offset+self.text_height, 0


        index = 0
        y = y_offset
        pli = 0
        for pl in pangolayouts_list:
            if y+self.text_height > text_area_height+self.top_margin:
                break
            pli += 1
            text_width = reduce(lambda a, b: a+b, map(lambda x: x[1], pl), 0)
            if len(pl) <= 1:
                # single word
                space = 0
            elif pli == len(pangolayouts_list):
                # last line in this paragraph
                space = self.space_width
            else:
                space = (text_area_width-text_width)/(len(pl)-1)
            x = x_offset
            for i in range(len(pl)):
                index = pl[i][2]
                if space and i == len(pl)-1 \
                       and not pli == len(pangolayouts_list):
                    # last word in this line
                    x = x_offset+text_area_width-pl[i][1]
                if pli == len(pangolayouts_list):
                    index = 0

                area_window.draw_layout(self.gc, x, y, pl[i][0],
                                        foreground=self.color)
                x += pl[i][1]+space

            y += self.text_height

        return y, index


    def get_pangolayout_list(self, line, text_area_width):
        '''create list:
        [
          [[pangolayout, width, index] [pangolayout, width, index] [...]]
          [[pangolayout, width, index] [pangolayout, width, index] [...]]
        ]
        '''

        words_list = []
        word = ''
        index = 0
        for c in line:
            index += 1
            if c in string.whitespace:
                if word:
                    words_list.append((word, index))
                    word = ''
                    continue
            word += c
        if word: words_list.append((word, index+1))

        text_width = 0
        total_width = 0
        line_number = 0
        pangolayout_list = []
        pangolayouts_list = []
        index = 0

        for word, index in words_list:

            pangolayout = self.create_pango_layout(word.encode('utf-8'))
            pangolayout.set_font_description(self.font)
            pangolayout.set_tabs(self.tab_array)
            width = pangolayout.get_extents()[1][2]/PANGO_SCALE

            if total_width+width > text_area_width:

                ## hyphanation
                wl = hyphenate(word)
                for w in wl:
                    pl =  self.create_pango_layout(w.encode('utf-8')+'-')
                    pl.set_font_description(self.font)
                    wd = pl.get_extents()[1][2]/PANGO_SCALE
                    if total_width+wd < text_area_width:
                        index -= len(word)-len(w)+1
                        pangolayout_list.append((pl, wd, index))
                        text_width += wd
                        # skip leading '-'
                        if word[len(w)] == '-':
                            wt = word[len(w)+1:]
                            index += 1
                        else:
                            wt = word[len(w):]
                        pangolayout = self.create_pango_layout(wt.encode('utf-8'))
                        pangolayout.set_font_description(self.font)
                        pangolayout.set_tabs(self.tab_array)
                        width = pangolayout.get_extents()[1][2]/PANGO_SCALE
                        break

                if pangolayout_list:
                    pangolayouts_list.append(pangolayout_list)

                line_number += 1
                text_width = 0
                total_width = 0
                pangolayout_list = []

            text_width += width
            total_width += width+self.space_width
            pangolayout_list.append((pangolayout, width, index))

        if pangolayout_list:
            pangolayouts_list.append(pangolayout_list)

        return pangolayouts_list


    def calc_back(self):

        def calc_back_1(line_index, index, text_area_width, text_area_height):
            new_line_index = line_index
            new_index = index
            y = 0
            if index:

                line = self.lines[line_index][:index]
                pangolayouts_list = self.get_pangolayout_list(line,
                                                              text_area_width)
                pangolayouts_list.reverse()

                if not pangolayouts_list:
                    # empty line
                    print 'empty line 1', new_line_index
                    y += self.text_height
                    if y+self.text_height > text_area_height:
                        return new_line_index-1, 0

                for pli in range(len(pangolayouts_list)):
                    pl = pangolayouts_list[pli]
                    y += self.text_height
                    if y+self.text_height > text_area_height:
                        new_index = pl[0][2]
                        if pli < len(pangolayouts_list)-1:
                            new_index = pangolayouts_list[pli+1][-1][2]
                        else:
                            new_index = 0

                        return new_line_index, new_index

            if y+self.text_height > text_area_height:
                print 'X'*24
                return new_line_index, 0

            lines = self.lines[:new_line_index]
            lines.reverse()

            for line in lines:

                new_line_index -= 1

                pangolayouts_list = self.get_pangolayout_list(line,
                                                              text_area_width)
                pangolayouts_list.reverse()

                if not pangolayouts_list:
                    # empty line
                    #print 'empty line 2', new_line_index
                    y += self.text_height
                    if y+self.text_height > text_area_height:
                        #print '*'*20
                        return new_line_index, 0
                    continue

                #import pprint
                #pprint.pprint(pangolayouts_list)
                for pli in range(len(pangolayouts_list)):
                    #pl = pangolayouts_list[pli+1]
                    y += self.text_height

                    if pli < len(pangolayouts_list)-1:
                        new_index = pangolayouts_list[pli+1][-1][2]
                    else:
                        new_index = 0

                    if y+self.text_height > text_area_height:

                        if pli < len(pangolayouts_list)-1:
                            new_index = pangolayouts_list[pli+1][-1][2]
                        else:
                            new_index = 0

                        return new_line_index, new_index


            return 0, 0

        text_area_height = self._height-self.top_margin-self.bottom_margin

        line_index = self.current_page_line_index
        index = self.current_page_char_index

        # right page
        x_offset = self._width/2+self.right_page_left_margin
        y_offset = self.top_margin
        text_area_width = self._width/2-self.right_page_left_margin \
                          -self.right_page_right_margin

        line_index, index = calc_back_1(line_index, index,
                                        text_area_width, text_area_height)

        # left page
        x_offset = self.left_page_left_margin
        y_offset = self.top_margin
        text_area_width = self._width/2-self.left_page_left_margin \
                          -self.left_page_right_margin

        line_index, index = calc_back_1(line_index, index,
                                        text_area_width, text_area_height)

        return line_index, index

##     def split_line(self, line):
##         words_list = []
##         word = ''
##         index = 0
##         prev_index = 0
##         for c in line:
##             index += 1
##             if c in string.whitespace:
##                 if word:
##                     words_list.append((word, prev_index))
##                     prev_index = index
##                     word = ''
##                     continue
##             word += c
##         if word: words_list.append((word, prev_index))
##         return words_list

    def go_fore_page(self):
##         print len(self.lines)
##         print self.current_page_line_index, self.current_page_char_index, \
##               self.next_page_line_index, self.next_page_char_index
        if self.next_page_line_index == len(self.lines):
            return
        self.current_page_line_index = self.next_page_line_index
        self.current_page_char_index = self.next_page_char_index
        self.draw_text()
        gtk.idle_add(self.emit_position_changed)


    def go_back_page(self):
        if self.current_page_line_index == 0 \
           and self.current_page_char_index == 0:
            return
        line_index, index = self.calc_back()
        self.current_page_line_index = line_index
        self.current_page_char_index = index
        self.draw_text()
        gtk.idle_add(self.emit_position_changed)


    def emit_position_changed(self):
        pos = float(reduce(lambda x, y: x+len(y),
                           self.lines[:self.current_page_line_index], 0) \
                    + self.current_page_char_index) / len(self.text)
        self.emit('position-changed', pos)


    def set_position(self, pos):
        index = len(self.text)*pos
        i = 0
        n = 0
        for line in self.lines:
            if index <= n:
                break
            n += len(line)+1
            i += 1
        self.current_page_line_index = i
        self.current_page_char_index = 0
        self.draw_text()


    def key_press_event_cb(self, w, event):
        print 'key_press_event_cb'

    def key_release_event_cb(self, w, event):
        #print 'key_release_event_cb'
        if event.keyval in self.keysyms_next:
            self.go_fore_page()
            return TRUE
        if event.keyval in self.keysyms_prev:
            self.go_back_page()
            return TRUE

    def button_release_event_cb(self, w, event):
        if event.button == 1:
            self.go_fore_page()
        elif event.button == 2:
            #print 'event.button == 2'
            self.go_back_page()


##----------------------------------------------------------------------

vowels = unicode('ţ', 'koi8-r')
consonants = unicode('', 'koi8-r')
hardsoftsigns = unicode('', 'koi8-r')

def hyphenate(word):
    # based on code by Mike Matsnev (Haali Reader)

    def hyphenate_word(word):

        length = len(word)

        i = 0
        wl = []
        w = ''
        while i < length:
            if word[i] in vowels:
                for j in range(i+1, length):
                    if word [j] in vowels:
                        if word[i+1] in consonants \
                               and word[i+2] in consonants:
                            w += word[i]
                            i += 1
                        elif word[i+1] in consonants \
                                 and word[i+2] in hardsoftsigns:
                            w += word[i:i+2]
                            i += 2
                        if i < length-2 and i >= 1:
                            ww = w+word[i]
                            wl.append(ww)
                        break

            w += word[i]
            i += 1

        return wl

    words_list = []
    word_split = word.split(u'-')
    w = u''
    ww = u''

    for i in word:
        if i.isalnum():
            ww += i
        else:
            if ww:
                for j in hyphenate_word(ww):
                    words_list.append(w+j)
            if i == '-':
                words_list.append(w+ww)
            w += ww+i
            ww = u''

    if ww:
        for j in hyphenate_word(ww):
            words_list.append(w+j)

    words_list.reverse()
    return words_list



if __name__ == '__main__':
    import sys
    if len(sys.argv) != 2:
        sys.exit('usage: %s file' % sys.argv[0])

    window = gtk.Window()
    window.connect('destroy', lambda w: gtk.main_quit())
    ob = OrnamentBook(window)
    window.add(ob)

    #text = ''
    #for i in xrange(20):
    #    text = text+str(i)*3+' '
    #text = (text+' aa bb cc ddddd\n')*7
    text = open(sys.argv[1]).read()
    ob.set_text(unicode(text, 'utf-8'))
    ob.realize()
    window.show_all()

    gtk.main()
